import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from furry_art_sync.sites.site import SiteProfile


class Datastore:
    STORE_DIR = "store"

    def __init__(self):
        os.makedirs(self.STORE_DIR, exist_ok=True)

    def save_profile(self, profile: "SiteProfile") -> None:
        pass  # TODO



import datetime
import json
import os
import re
from pathlib import Path
from typing import List, Dict, Optional, Type

import gallery_dl
import requests

from furry_art_sync.datastore import Datastore
from furry_art_sync.sites.post import Post, PostRating
from furry_art_sync.sites.site import SiteProfile, SiteUploader


class WeasylPost(Post):
    @property
    def link(self) -> str:
        return self.metadata_raw["link"]

    @property
    def title(self) -> Optional[str]:
        return self.metadata_raw["title"]

    @property
    def description(self) -> Optional[str]:
        return self.metadata_raw["description"]

    @property
    def tags(self) -> Optional[List[str]]:
        return self.metadata_raw["tags"]

    @property
    def rating(self) -> Optional[PostRating]:
        return {
            "general": PostRating.SAFE,
            "mature": PostRating.MATURE,
            "explicit": PostRating.EXPLICIT,
        }[self.metadata_raw["rating"].lower()]

    @property
    def datetime_posted(self) -> Optional[datetime.datetime]:
        return datetime.datetime.fromisoformat(self.metadata_raw["posted_at"])


class WeasylUploader(SiteUploader):

    def __init__(self, profile: "WeasylSiteProfile") -> None:
        self.profile = profile
        self.api_key = profile.api_key

    @classmethod
    def user_setup_uploader(cls, site_profile: "WeasylSiteProfile") -> "WeasylUploader":
        return cls(site_profile)

    def upload_post(self, post: "Post") -> "Post":
        upload_rating = {
            PostRating.SAFE: 10,
            PostRating.MATURE: 30,
            PostRating.EXPLICIT: 40,
        }.get(post.rating, 40)
        tags = post.tags or []
        tags_modified = False
        while len(tags) < 2:
            print("ERROR: Weasyl requires at least 2 tags on submissions.")
            new_tags = input("Please enter at least 2 tags, separated by commas")
            tags = [tag.strip().replace(" ", "_") for tag in new_tags.split(",")]
            tags_modified = True
        if tags_modified:
            print(f"Tags have been set to {tags} for Weasyl upload")
        data = {
            "title": post.title,
            "rating": upload_rating,
            "content": post.description,
            "tags": " ".join(tags),
        }
        files = {
            "submitfile": open(post.file_path, "rb"),
        }
        post_type = {
            "jpg": "visual",
            "jpeg": "visual",
            "png": "visual",
            "gif": "visual",
            "pdf": "literary",
            "txt": "literary",
            "mp3": "multimedia",
            "swf": "multimedia",
        }.get(post.file_ext)
        if post_type is None:
            raise ValueError(f"This post file type, \".{post.file_ext}\" is not supported on Weasyl")
        url = f"https://weasyl.com/submit/{post_type}"
        if post_type in ["literary", "multimedia"]:
            files["coverfile"] = None  # TODO
        resp = requests.post(
            url,
            data=data,
            files=files,
            headers={
                "Referer": url,
                "Origin": "https://www.weasyl.com",
                "Host": "www.weasyl.com",
                "X-Weasyl-API-Key": self.api_key,
              },
        )
        if resp.status_code != 200:
            raise ValueError("Failed to upload to Weasyl")
        new_url = resp.url
        new_id = re.compile("submissions/([0-9]+)/").search(new_url).group(1)
        new_data = requests.get(
            f"https://www.weasyl.com/api/submissions/{new_id}/view",
            headers={
                "X-Weasyl-API-Key": self.api_key
            }
        ).json()
        os.makedirs(self.profile.profile_directory(), exist_ok=True)
        data_path = self.profile.profile_directory() / f"{new_id}.json"
        with open(data_path, "w") as f:
            json.dump(new_data, f, indent=2)
        content_url = new_data["media"]["submission"][0]["url"]
        content_resp = requests.get(content_url)
        new_ext = new_data["media"]["submission"][0]["url"].split(".")[-1].lower()
        file_path = self.profile.profile_directory() / f"{new_id}.{new_ext}"
        with open(file_path, "wb") as f:
            f.write(content_resp.content)
        return WeasylPost(file_path, data_path, new_data, self.profile)


class WeasylSiteProfile(SiteProfile):
    SITE_DIR = "weasyl"

    def __init__(self, username: str, api_key: str) -> None:
        self.username = username
        self.api_key = api_key

    def profile_directory(self) -> Path:
        return Path(Datastore.STORE_DIR) / self.SITE_DIR / self.username

    def profile_link(self) -> str:
        return f"https://weasyl.com/~{self.username}"

    def validate(self) -> bool:
        resp = requests.get(self.profile_link())
        if resp.status_code == 404:
            print("Error: Weasyl user does not exist")
            return False
        api_resp = requests.get(
            "https://weasyl.com/api/whoami",
            headers={
                "X-Weasyl-API-Key": self.api_key
            },
        )
        if api_resp.status_code != 200:
            print("Error: Failure authenticating to Weasyl API")
            return False
        api_data = api_resp.json()
        if "error" in api_data:
            print(f"Error: Weasyl API returned error: {api_data['error']}")
            return False
        return True

    def configure_gallery_dl(self) -> None:
        gallery_dl.config.set(("extractor",), "filename", "{submitid}.{extension}")
        gallery_dl.config.set(("extractor", "weasyl"), "metadata", True)
        gallery_dl.config.set(("extractor", "weasyl"), "api-key", self.api_key)

    def gallery_dl_postprocessors(self) -> List[Dict[str, str]]:
        return [
            {
                "name": "metadata",
                "mode": "json",
                "filename": "{submitid}.json",
            }
        ]

    def build_post(self, submission_file: Path, metadata_file: Path, post_metadata: Dict) -> Post:
        return WeasylPost(
            submission_file,
            metadata_file,
            post_metadata,
            self,
        )

    @classmethod
    def user_setup_profile(cls) -> "SiteProfile":
        raw_username = input("Please enter your Weasyl username: ")
        username = raw_username.lower()
        api_key = input("Please go to https://www.weasyl.com/control/apikeys and generate an API key: ")
        return cls(username, api_key)

    def uploader_class(self) -> Optional[Type["SiteUploader"]]:
        return WeasylUploader



import datetime
import enum
import hashlib
import os.path
from abc import abstractmethod, ABC
from functools import cached_property

import imagehash
from PIL import Image
from pathlib import Path
from typing import Optional, List, Dict, TYPE_CHECKING

from furry_art_sync.sites.post_match import PostMatch, MD5Match, HDImageHashMatch, LowDetailImageHashMatch

if TYPE_CHECKING:
    from furry_art_sync.sites.site import SiteProfile


class PostRating(enum.Enum):
    SAFE = "safe"
    MATURE = "mature"
    EXPLICIT = "explicit"


class Post(ABC):
    EXTENSIONS_IMAGE = ["png", "gif", "jpg", "jpeg"]
    EXTENSIONS_ANIM = ["gif", "png"]
    EXTENSIONS_VIDEO = ["webm", "mp4"]

    def __init__(self, file_path: Path, metadata_path: Path, metadata_raw: Dict, site_profile: "SiteProfile") -> None:
        self.file_path = file_path
        self.metadata_path = metadata_path
        self.metadata_raw = metadata_raw
        self.site_profile = site_profile
        self._md5_hash: Optional[str] = None

    @property
    @abstractmethod
    def link(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def title(self) -> Optional[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def description(self) -> Optional[str]:
        raise NotImplementedError

    @property
    @abstractmethod
    def tags(self) -> Optional[List[str]]:
        raise NotImplementedError

    @property
    @abstractmethod
    def rating(self) -> Optional[PostRating]:
        raise NotImplementedError

    @property
    @abstractmethod
    def datetime_posted(self) -> Optional[datetime.datetime]:
        raise NotImplementedError

    @property
    def file_ext(self) -> str:
        _, ext = os.path.splitext(self.file_path)
        return ext.lower().removeprefix(".")

    @cached_property
    def is_static_image(self) -> bool:
        if self.file_ext.lower() not in self.EXTENSIONS_IMAGE:
            return False
        with Image.open(self.file_path) as img:
            return not getattr(img, "is_animated", False)

    @cached_property
    def colour_hash(self) -> Optional[imagehash.ImageHash]:
        if not self.is_static_image:
            return None
        with Image.open(self.file_path) as img:
            return imagehash.colorhash(img)

    @cached_property
    def high_fidelity_phash(self) -> Optional[imagehash.ImageHash]:
        if not self.is_static_image:
            return None
        with Image.open(self.file_path) as img:
            return imagehash.phash(img, hash_size=32)

    @cached_property
    def low_fidelity_phash(self) -> Optional[imagehash.ImageHash]:
        if not self.is_static_image:
            return None
        with Image.open(self.file_path) as img:
            return imagehash.phash(img, hash_size=8)

    @cached_property
    def md5_hash(self) -> str:
        with open(self.file_path, "rb") as f:
            return hashlib.md5(f.read()).hexdigest()

    def matches_post(self, other: "Post") -> Optional[PostMatch]:
        if self.md5_hash == other.md5_hash:
            return MD5Match()
        if self.is_static_image and other.is_static_image:
            if self.colour_hash - other.colour_hash < 2:
                if self.low_fidelity_phash - other.low_fidelity_phash <= 2:
                    if self.high_fidelity_phash - other.high_fidelity_phash <= 4:
                        return HDImageHashMatch()
                    return LowDetailImageHashMatch()
        pass  # TODO: Implement for gifs?

    def matches_any_posts(self, others: List["Post"]) -> Dict["Post", PostMatch]:
        return {
            post: self.matches_post(post) for post in others
        }



import glob
import json
import os.path
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Dict, TYPE_CHECKING, Optional, Type

import gallery_dl

from furry_art_sync.datastore import Datastore

if TYPE_CHECKING:
    from furry_art_sync.sites.post import Post


class SiteProfile(ABC):

    @abstractmethod
    def validate(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def build_post(self, submission_file: Path, metadata_file: Path, post_metadata: Dict) -> "Post":
        raise NotImplementedError

    def list_local_posts(self) -> List["Post"]:
        metadata_files = glob.glob("**/*.json", root_dir=self.profile_directory(), recursive=True)
        posts = []
        for metadata_file in metadata_files:
            metadata_path = self.profile_directory() / metadata_file
            with open(metadata_path, "r") as f:
                metadata = json.load(f)
            submission_pattern = metadata_file.removesuffix("json") + "*"
            submission_file = [
                s for s in glob.glob(submission_pattern, root_dir=self.profile_directory()) if not s.endswith(".json")
            ][0]
            submission_path = self.profile_directory() / submission_file
            posts.append(self.build_post(submission_path, metadata_path, metadata))
        return posts

    def configure_gallery_dl(self) -> None:
        pass  # Sites can optionally override this, to set other gallery dl config

    def gallery_dl_postprocessors(self) -> List[Dict[str, str]]:
        return [
            {
                "name": "metadata",
                "mode": "json",
                "filename": "{id}.json",
            }
        ]

    def download_posts(self) -> None:
        config_path = str(Path(Datastore.STORE_DIR) / "gallery_dl_config.json")
        if not os.path.exists(config_path):
            with open(config_path, "w") as f:
                json.dump({}, f)
        gallery_dl.config.load([config_path], strict=True)
        gallery_dl.config.set(("extractor",), "base-directory", ".")
        gallery_dl.config.set(("extractor",), "directory", self.profile_directory().parts)
        gallery_dl.config.set(("extractor",), "filename", "{id}.{extension}")
        gallery_dl.config.set(("extractor",), "postprocessors", self.gallery_dl_postprocessors())
        self.configure_gallery_dl()
        data_job = gallery_dl.job.DataJob(self.profile_link())
        dl_job = gallery_dl.job.DownloadJob(self.profile_link())
        data_resp = data_job.run()
        dl_resp = dl_job.run()
        print(data_resp)
        print(dl_resp)

    @abstractmethod
    def profile_directory(self) -> Path:
        raise NotImplementedError

    @abstractmethod
    def profile_link(self) -> str:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def user_setup_profile(cls) -> "SiteProfile":
        raise NotImplementedError

    def uploader_class(self) -> Optional[Type["SiteUploader"]]:
        return None


class SiteUploader(ABC):

    @classmethod
    @abstractmethod
    def user_setup_uploader(cls, site_profile: SiteProfile) -> "SiteUploader":
        raise NotImplementedError

    @abstractmethod
    def upload_post(self, post: "Post") -> "Post":
        raise NotImplementedError



import json
import os
import time
import datetime
from pathlib import Path
from typing import Optional, List, Dict

import requests

from furry_art_sync.datastore import Datastore
from furry_art_sync.sites.post import Post, PostRating
from furry_art_sync.sites.site import SiteProfile


class FurAffinityPost(Post):
    @property
    def link(self) -> str:
        return self.metadata_raw["link"]

    @property
    def title(self) -> Optional[str]:
        return self.metadata_raw["title"]

    @property
    def description(self) -> Optional[str]:
        return self.metadata_raw["description"]

    @property
    def tags(self) -> Optional[List[str]]:
        return self.metadata_raw["keywords"]

    @property
    def rating(self) -> Optional[PostRating]:
        return {
            "general": PostRating.SAFE,
            "mature": PostRating.MATURE,
            "adult": PostRating.EXPLICIT,
        }[self.metadata_raw["rating"].lower()]

    @property
    def datetime_posted(self) -> Optional[datetime.datetime]:
        return datetime.datetime.fromisoformat(self.metadata_raw["posted_at"])


class FurAffinitySiteProfile(SiteProfile):
    SITE_DIR = "furaffinity"

    def __init__(self, username: str) -> None:
        self.username = username

    def profile_link(self) -> str:
        return f"https://www.furaffinity.net/user/{self.username}"

    def profile_directory(self) -> Path:
        return Path(Datastore.STORE_DIR) / self.SITE_DIR / self.username

    @classmethod
    def user_setup_profile(cls) -> "SiteProfile":
        raw_username = input("Please enter your FurAffinity username: ")
        username = raw_username.lower().replace("_", "")
        return cls(username)

    def validate(self) -> bool:
        resp = self._request_from_api(f"/user/{self.username}.json")
        if resp.status_code == 404:
            return False
        return True

    def download_posts(self) -> None:
        page = 1
        while self._download_page("gallery", page):
            page += 1
        page = 1
        while self._download_page("scraps", page):
            page += 1
        print("Download complete")

    def _download_page(self, folder: str, page: int) -> bool:
        print(f"Downloading page {page} of FA {folder} for {self.username}")
        directory = self.profile_directory()
        if folder == "scraps":
            directory = directory / "Scraps"
        os.makedirs(directory, exist_ok=True)
        resp = self._request_from_api(f"/user/{self.username}/{folder}.json?full=1&page={page}")
        resp_json = resp.json()
        if not resp_json:
            return False
        for post_data in resp.json():
            post_id = post_data["id"]
            print(f"Downloading post {post_id}")
            metadata_path = directory / f"{post_id}.json"
            if os.path.exists(metadata_path):
                continue
            full_post_resp = self._request_from_api(f"/submission/{post_id}.json")
            full_post_data = full_post_resp.json()
            with open(metadata_path, "w") as f:
                json.dump(full_post_data, f, indent=2)
            image_url = full_post_data["download"]
            image_ext = image_url.split(".")[-1]
            self._download_file(image_url, directory / f"{post_id}.{image_ext}")
        return True

    def _request_from_api(self, path: str) -> requests.Response:
        max_attempts = 10
        attempt = 0
        resp = None
        while attempt < max_attempts:
            resp = requests.get(f"https://faexport.spangle.org.uk/{path.lstrip('/')}")
            if resp.status_code == 503:
                # Cloudflare error
                time.sleep(attempt * 0.5)
                continue
            break
        if resp is None:
            raise Exception("Request did not get made to FAExport API")
        if resp.status_code == 503:
            raise Exception("FA seems to be under cloudflare protection at the moment")
        return resp

    def _download_file(self, image_url: str, target_path: Path) -> None:
        if os.path.exists(target_path):
            return
        resp = requests.get(image_url)
        image_content = resp.content
        with open(target_path, "wb") as f:
            f.write(image_content)

    def build_post(self, submission_file: Path, metadata_file: Path, post_metadata: Dict) -> Post:
        return FurAffinityPost(
            submission_file,
            metadata_file,
            post_metadata,
            self,
        )



from typing import List, Type


class PostMatch:
    def __init__(self, match_type: str):
        self.match_type = match_type

    def priority(self) -> int:
        return HASH_PRIORITY.index(self.__class__)


class ManualMatch(PostMatch):
    def __init__(self):
        super().__init__("manual")


class MD5Match(PostMatch):
    def __init__(self):
        super().__init__("MD5 hashes match")


class HDImageHashMatch(PostMatch):
    def __init__(self):
        super().__init__("HD image hashes match")


class LowDetailImageHashMatch(PostMatch):
    def __init__(self):
        super().__init__("Low detail image hashes match")


HASH_PRIORITY: List[Type[PostMatch]] = [ManualMatch, MD5Match, HDImageHashMatch, LowDetailImageHashMatch]



import os.path
from functools import cache
from typing import Dict, List

from furry_art_sync.datastore import Datastore
from furry_art_sync.sites.post import Post
from furry_art_sync.sites.post_match import PostMatch, HASH_PRIORITY
from furry_art_sync.sites.site import SiteProfile
from furry_art_sync.sites.site_weasyl import WeasylSiteProfile
from furry_art_sync.sites.site_furaffinity import FurAffinitySiteProfile


class Tool:
    SUPPORTED_SITES = {
        "furaffinity": FurAffinitySiteProfile,
        "weasyl": WeasylSiteProfile,
    }

    def __init__(self):
        self.datastore = Datastore()

    def run(self) -> None:
        print("Welcome to Furry Art Sync. A tool to synchronise your gallery across multiple furry art sites")
        while True:
            print("----")
            print("There are a few things you can do with this tool:")
            print("[1] Download a new site profile")
            print("[2] Check the match between two site profiles")
            print("[q] Quit the application")
            choice = input("What would you like to do? Enter a number to select: ")
            if choice == "1":
                self.download_profile()
            elif choice == "2":
                self.check_match()
            elif choice.lower() in ["q", "quit", "exit"]:
                print("Thank you for using Furry Art Sync")
                break
            else:
                print("I did not understand that choice, sorry.")

    def download_profile(self) -> None:
        while True:
            print("Which site would you like to set up a profile on?")
            site_idx = {str(n + 1): site_name for n, site_name in enumerate(self.SUPPORTED_SITES.keys())}
            for n, site_name in site_idx.items():
                print(f"- [{n}] {site_name}")
            choice = input("Please select a site by entering the number of the site: ")
            if choice.lower() in site_idx.keys():
                site_class = self.SUPPORTED_SITES[site_idx[choice.lower()]]
                break
            elif choice.lower() in self.SUPPORTED_SITES.keys():
                site_class = self.SUPPORTED_SITES[choice.lower()]
                break
            else:
                print("Choice was not recognised.")
        # Set up profile
        while True:
            profile: SiteProfile = site_class.user_setup_profile()
            valid = profile.validate()
            if not valid:
                print("This does not seem to be a valid profile")
                continue
            break
        self.datastore.save_profile(profile)
        print("Profile added")
        print("Downloading profile")
        profile.download_posts()
        self.datastore.save_profile(profile)

    def check_match(self) -> None:
        print("First, let's set up your FA data")
        while True:
            fa_profile: SiteProfile = FurAffinitySiteProfile.user_setup_profile()
            valid = fa_profile.validate()
            if not valid:
                print("This does not seem to be a valid FA profile")
                continue
            break
        print("Okay, now setup your weasyl data")
        while True:
            weasyl_profile: SiteProfile = WeasylSiteProfile.user_setup_profile()
            valid = weasyl_profile.validate()
            if not valid:
                print("This does not seem to be a valid Weasyl profile")
                continue
            break
        comparison = ProfileComparison(fa_profile, weasyl_profile)
        print(f"You have {len(comparison.source_posts())} posts on FA")
        print(f"You have {len(comparison.target_posts())} posts on Weasyl")
        for fa_post, match_dict in comparison.matched_posts().items():
            if match_dict:
                print(f"MATCH: {fa_post.link} matches {len(match_dict)} posts on weasyl:")
                for weasyl_post, match in match_dict.items():
                    print(f"- {weasyl_post.link}: {match.match_type}")
            else:
                print(f"NO MATCH: {fa_post.link}")
        print("Unmatched posts from source: ")
        for fa_post in comparison.unmatched_source_posts():
            print(f"- {fa_post.link} \"{fa_post.title}\"")
        print("Unmatched posts from target: ")
        for weasyl_post in comparison.unmatched_target_posts():
            print(f"- {weasyl_post.link} \"{weasyl_post.title}\"")
        print("---")
        weasyl_upload_class = weasyl_profile.uploader_class()
        if not weasyl_upload_class:
            print("No uploader is configured for this target, so cannot automatically sync your gallery")
            return
        upload_option_resp = input("Would you like to upload missing posts to Weasyl? [yN] ")
        if upload_option_resp.lower() not in ["y", "yes"]:
            print("Okay, we will not sync your gallery")
            return
        weasyl_uploader = weasyl_upload_class.user_setup_uploader(weasyl_profile)
        fa_unmatched = sorted(comparison.unmatched_source_posts(), key=lambda post: post.datetime_posted)
        for fa_post in fa_unmatched:
            print(f"This post exists on FA, but not Weasyl: {fa_post.link}, \"{fa_post.title}\"")
            upload_post_resp = input("Would you like to upload that post to Weasyl? [yN]")
            if upload_post_resp.lower() in ["y", "yes"]:
                weasyl_post = weasyl_uploader.upload_post(fa_post)
                print(f"Post uploaded: {weasyl_post.link}")
            print("--")
        print("All done!")


class ProfileComparison:
    def __init__(self, source: SiteProfile, target: SiteProfile) -> None:
        self.source = source
        self.target = target
        self.has_downloaded = False

    def download(self) -> None:
        if self.has_downloaded:
            return
        self.source.download_posts()
        self.target.download_posts()

    @cache
    def source_posts(self) -> List[Post]:
        self.download()
        return self.source.list_local_posts()

    @cache
    def target_posts(self) -> List[Post]:
        self.download()
        return self.target.list_local_posts()

    @cache
    def matched_posts(self) -> Dict[Post, Dict[Post, PostMatch]]:
        results = {}
        source_posts = self.source_posts()
        target_posts = self.target_posts()
        for source_post in sorted(source_posts, key=lambda post: post.link):
            match_dict = source_post.matches_any_posts(target_posts)
            match_dict_filtered: Dict[Post, PostMatch] = {
                post: match for post, match in match_dict.items() if match is not None
            }
            if not match_dict_filtered:
                results[source_post] = {}
                continue
            best_match_priority = min(match.priority() for match in match_dict_filtered.values())
            best_matches = {
                post: match for post, match in match_dict_filtered.items() if match.priority() == best_match_priority
            }
            # TODO: Ensure each target post is only matched to once?
            results[source_post] = best_matches
            if best_matches:
                print(f"MATCH: {source_post.link} matches {len(best_matches)} posts on target:")
                for target_post, match in best_matches.items():
                    print(f"- {target_post.link}: {match.match_type}")
            else:
                print(f"NO MATCH: {source_post.link}")
        return results

    def run(self) -> None:
        pass

    @cache
    def unmatched_source_posts(self) -> List[Post]:
        matched_posts = self.matched_posts()
        return [
            post for post, matches in matched_posts.items() if not matches
        ]

    @cache
    def unmatched_target_posts(self) -> List[Post]:
        target_posts = set(self.target_posts())
        for matches in self.matched_posts().values():
            target_posts -= set(matches.keys())
        return list(target_posts)
